#
#
#
#
#   peer, port,
#
#
#
package Cleo::Conn;

use strict;
use Exporter;

use IO::Socket::INET 1.29;
use IO::Select;
use Tie::RefHash;
use POSIX;

use vars qw($VERSION @ISA @EXPORT $conn_timeout);

#
#  States: dead ---> connecing -> ok
#           ^    \-> listen ->-/  |
#           \---------------------/
#
#
#

#my $_d=0;
#sub logit{
#    print STDERR $_[0] if(_d);
#}

my %_conn_by_h;
tie %_conn_by_h, "Tie::RefHash";

# conn, IO::Socket
sub _add_conn{
    $_conn_by_h{$_[1]}=$_[0];
#!    print "ADD: {$_[1]}=$_[0]\n;";
}

# IO::Socket
sub _del_conn{
    delete $_conn_by_h{$_[0]};
#!    print "DEL: $_[0]\n;";
}

$VERSION=1.0;
@ISA=();
@EXPORT = qw(new new_handle new_listen connect disconnect get_peer get_port
             set_port set_peer get_state get_h get_conn set_timeout
             listen accept send read unread flush allflush);

    # constructor
    # opt args: peer, port
    sub new {
        my $self ={};
        $self->{peer} = $_[1];
        $self->{port} = $_[2];
        $self->{state}='dead';
        $self->{handle}=undef;
        $self->{sbuffer}='';
        $self->{rbuffer}='';
        $self->{conn_timeout}=60;

        bless($self);

        return $self;
    }

    # constructor
    # args: port [num listens]
    sub new_listen {
        my $self ={};
        undef $self->{peer};
        $self->{port} = $_[1];
        $self->{listen} = $_[2];
        $self->{state}='dead';
        $self->{handle}=undef;
        $self->{sbuffer}='';
        $self->{rbuffer}='';

        bless($self);
        return $self;
    }

    # constructor
    # args: IO::Handle
    sub new_handle {
        my $self ={};
        undef $self->{peer};
        $self->{file} = 1;
        $self->{state}='ok';
        $self->{handle}=$_[1];
        $self->{sbuffer}='';
        $self->{rbuffer}='';

        _add_conn($self,$_[1]);
        bless($self);
        return $self;
    }

    sub DESTROY{
        my $self=shift;
        if(defined $self->{handle}){
            _del_conn($self->{handle});
            $self->{handle}->flush;
            $self->{handle}->close;
        }
    }

    # set peername and port
    sub set_peer{
        $_[0]->{peer}=$_[1];
    }
    sub set_port{
        $_[0]->{port}=$_[1];
    }

    sub set_timeout{
        $_[0]->{conn_timeout}=$_[1];
    }
    # get peername and port
    sub get_peer{
        return $_[0]->{peer};
    }
    sub get_port{
        return $_[0]->{port};
    }

    # - dead
    # - ok
    # - connecting
    # - listen
    #
    sub get_state{
        my $self=shift;

        if($self->{state} eq 'connecting'){
            if($self->{handle}->connected){
                $self->{state} = 'ok';
            }
            # timed out?
            elsif($self->{conn_time}+$self->{conn_timeout}<time()){
                # still connecting...
                _del_conn($self);
                $self->{handle}->close;
                delete $self->{handle};
                $self->{state}='dead';
            }

        }
        return $self->{state};
    }

    # try to connect
    #
    # !!!!  NOTE  !!!!
    # handle can be changed!
    #
    #
    #
    sub connect{
        my $self=shift;

        return 0 if($self->{state} eq 'ok');

        # already connecting?
        if(defined $self->{handle}){
            if($self->{handle}->connected){
                $self->{state} = 'ok';
                return 0;
            }
            # timed out?
            if($self->{conn_time}+$self->{conn_timeout}>time()){
                # still connecting...
                return 1;
            }
            _del_conn_($self->{handle});
            $self->{handle}->close;
            undef $self->{handle};
        }

        # new connect attempt
        $self->{conn_time}=time();
        $self->{handle} = IO::Socket::INET->new(
            PeerAddr => $self->{peer},
            PeerPort => $self->{port},
            Proto    => 'tcp',
            Reuse => 1,
            Blocking => 0 );

        if(defined $self->{handle}){
            _add_conn($self,$self->{handle});
            $self->{state}= 'connecting';
            $self->{handle}->blocking(0);
            return 0 ;
        }
        else{
            #!print "Cannot connect to $self->{peer}:$self->{port} ($!)\n";
        }
        return 1;
    }

    #
    # optional arg: do NOT flush data before closing connection.
    #
    sub disconnect{
        my $self=shift;
        my $no_flush=shift;

        return 0 if($self->{state} eq 'dead');

        $self->flush unless $no_flush;
        _del_conn($self->{handle});
        $self->{handle}->close;
        undef $self->{handle};
        $self->{state}='dead';
    }

    #
    #  optional argument - adderss binding
    #
    sub listen{
        my $self=shift;
        my $addr=shift;

        $addr |= '0.0.0.0';

        if(defined $self->{handle}){
            _del_conn($self->{handle});
            $self->{handle}->close;
        }
        $self->{listen} |= 1;
        $self->{handle}=IO::Socket::INET->new(
                Listen     => $self->{listen},
                LocalPort  => $self->{port},
                LocalAddr  => $addr,
                Proto      => 'tcp',
                ReuseAddr  => 1,
                #ReusePort  => 1,
                Blocking   => 0 );
        return 1 unless defined $self->{handle};

        _add_conn($self,$self->{handle});
        $self->{state}='listen';
        return 0;
    }

    sub accept{
        my $self=shift;

        return undef unless $self->{state} eq 'listen';
        my $h=$self->{handle}->accept();

        if(defined $h){
            my $new_conn={};

            $h->blocking(0);
            $new_conn->{peer} = $h->peerhost;
            $new_conn->{port} = $h->peerport;
            $new_conn->{state}='ok';
            $new_conn->{handle}=$h;
            $new_conn->{sbuffer}='';
            $new_conn->{rbuffer}='';
            $new_conn->{conn_timeout}=60;

            bless($new_conn);
            _add_conn($new_conn,$h);

            return $new_conn;
        }
        return undef;
    }

    # push to send-buffer
    sub send{
        my $self=shift;
        $self->{sbuffer} .= $_[0];
    }

    #actually send
    sub flush{
        my $self=shift;
        my ($res,$len);

        $len=length $self->{sbuffer};
        return 0 if $len==0;

        return 1 unless $self->{state} eq 'ok';

        #try to send
        $res = $self->{handle}->syswrite($self->{sbuffer});
        if(!defined $res){
            # ERROR
            return 1 if(!defined $res and ($! == EAGAIN));
            if($self->{file}){
                _del_conn($self->{handle});
                $self->{state}='dead';
                $self->{handle}->close;
                undef $self->{handle};
                return 1;
            }

            # it is a socket. try to reconnect later
            $self->disconnect(1);
            if($self->{listen}){
                $self->listen();
                return 1;
            }
            return 1;
        }
        # sent
        substr($self->{sbuffer},0,$res)='';
    }

    sub read{
        my $self=shift;
        my ($res,$readed);

        if($self->get_state() ne 'ok'){
            $readed=$self->{rbuffer};
            $self->{rbuffer}='';
            return $readed if $readed ne '';
            return undef;
        }

        $res = $self->{handle}->sysread($readed,4096);
#        logit("SYSREAD=$readed.\n");
        if((!defined $res) or ($res==0)){
            # ERROR or EOF
            if(!defined $res and ($! == EAGAIN)){
                # nothing was readed
                $readed=$self->{rbuffer};
                $self->{rbuffer}='';
                return $readed;
            }
            if($self->{file}){
                _del_conn($self->{handle});
                $self->{handle}->close;
                undef $self->{handle};
                $self->{state}='dead';

                $readed=$self->{rbuffer};
                $self->{rbuffer}='';
                return $readed if $readed ne '';
                return undef;
            }

            # it is a socket. try to reconnect later
            $self->disconnect;
            $readed=$self->{rbuffer};
            $self->{rbuffer}='';
            return $readed if $readed ne '';
            return undef;
        }

        $readed=$self->{rbuffer}.$readed;
        $self->{rbuffer}='';
        return $readed;
    }

    sub unread{
        my $self=shift;
        $self->{rbuffer}=$_[0];
    }

    sub get_h{
        return $_[0]->{handle};
    }

    sub get_conn{
        shift if($_[0] eq 'Cleo::Conn');
        if(defined $_conn_by_h{$_[0]}){
            return $_conn_by_h{$_[0]};
        }
        return undef;
    }

    sub allflush{
        foreach my $i (keys(%_conn_by_h)){
            $_conn_by_h{$i}->flush();
        }
    }
1;

__END__

=HEAD1 EXAMPLE

use IO::Select;
use Cleo::Conn;

$x=new Cleo::Conn("localhost",2525);
$x->set_timeout(5);
$x->connect;

while(($st=$x->get_state) ne 'ok'){
  print "$st\n";
  if($st eq 'dead'){
    $x->connect;
  }
  sleep 2;
}
print "Connected\n";
$s=new IO::Select($x->get_h);

do {
  @a=$s->can_read;
  sleep 1;
}while @a<1;

$q=Cleo::Conn->get_conn($a[0]);
if(defined $q){
  print $q->get_peer()."\n";
}
else{
  print "ERROR\n";
}
